from ._fast_stats_ext import (
    _binary_f1_score_reqs,
    _binary_precision_reqs,
    _binary_recall_reqs,
    _confusion_matrix,
    _unique,
)
from .binary import binary_f1_score, binary_precision, binary_recall
from .confusion_matrix import confusion_matrix
